# pytest-steps

*Create step-wise / incremental tests in `pytest`.*

[![Build Status](https://travis-ci.org/smarie/python-pytest-steps.svg?branch=master)](https://travis-ci.org/smarie/python-pytest-steps) [![Tests Status](https://smarie.github.io/python-pytest-steps/junit/junit-badge.svg?dummy=8484744)](https://smarie.github.io/python-pytest-steps/junit/report.html) [![codecov](https://codecov.io/gh/smarie/python-pytest-steps/branch/master/graph/badge.svg)](https://codecov.io/gh/smarie/python-pytest-steps) [![Documentation](https://img.shields.io/badge/docs-latest-blue.svg)](https://smarie.github.io/python-pytest-steps/) [![PyPI](https://img.shields.io/badge/PyPI-pytest_steps-blue.svg)](https://pypi.python.org/pypi/pytest_steps/)

Did you ever want to organize your test is incremental steps, for example to improve readability in case of failure ? Or to have some optional steps, executing only conditionally to previous steps' results?

`pytest-steps` leverages `pytest` and its great `@pytest.mark.parametrize` and `@pytest.fixture` decorators, so that you can **create incremental tests with steps** without having to think about the pytest fixture/parametrize pattern that has to be implemented for your particular case. 

This is particularly helpful if:
 
 - you wish to **share a state / intermediate results** across test steps
 - your tests **already rely on other fixtures and/or parameters**, such as with [pytest-cases](https://smarie.github.io/python-pytest-cases/). In that case, finding the correct pytest design, that will ensure that you have a brand new state object at the beginning of each test suite while ensuring that this object will actually be shared across all steps, might be very tricky. 

With `pytest-steps` you don't have to care about the internals: it just works as expected.


## Installing

```bash
> pip install pytest_steps
```

## Usage
 
Simply decorate your test function with `@test_steps`:

```python
from pytest_steps import test_steps

def step_a():
    # perform this step ...
    print("step a")
    assert not False  # replace with your logic

def step_b():
    # perform this step
    print("step b")
    assert not False  # replace with your logic

@test_steps(step_a, step_b)
def test_suite_no_shared_results(test_step):
    # Execute the step
    test_step()
```

You can add a `steps_data` parameter to your test function if you wish to share a `StepsDataHolder` object between your steps.

```python
import pytest
from pytest_steps import test_steps, StepsDataHolder

def step_a(steps_data):
    # perform this step ...
    print("step a")
    assert not False  # replace with your logic

    # intermediate results can be stored in steps_data
    steps_data.intermediate_a = 'some intermediate result created in step a'

def step_b(steps_data):
    # perform this step, leveraging the previous step's results
    print("step b")
    
    # you can leverage the results from previous steps... 
    # ... or pytest.skip if not relevant
    if len(steps_data.intermediate_a) < 5:
        pytest.skip("Step b should only be executed if the text is long enough")
    
    new_text = steps_data.intermediate_a + " ... augmented"  
    print(new_text)
    assert len(new_text) == 56

@test_steps(step_a, step_b)
def test_suite_with_shared_results(test_step, steps_data: StepsDataHolder):

    # Execute the step with access to the steps_data holder
    test_step(steps_data)
```

You can add as many `@pytest.mark.parametrize` and pytest fixtures in your test suite function, it should work as expected: the steps_data object will be created everytime a new parameter/fixture combination is created, but will be shared across steps with the same parameters and fixtures.

You can also check the [API Reference](./api_reference.md).


## Main features / benefits

 * **Split tests into steps**. Although the best practices in testing are very much in favor of having each test completely independent of the other ones (for example for distributed execution), there is definitely some value in results readability to break down tests into chained sub-tests (steps). The `@test_steps` decorator provides an intuitive way to do that without forcing any data model (steps can be functions, objects, etc.).
 * **Optional shared placeholder**: All steps in the same test share the same `StepsDataHolder` object. You simply have to add a `steps_data` parameter to your test function (name configurable).


## See Also

 - [pytest documentation on parametrize](https://docs.pytest.org/en/latest/parametrize.html)
 - [pytest documentation on fixtures](https://docs.pytest.org/en/latest/fixture.html)
 - [pytest-cases](https://smarie.github.io/python-pytest-cases/), to go further and separate test data from test functions

### Others

*Do you like this library ? You might also like [my other python libraries](https://github.com/smarie/OVERVIEW#python)* 

## Want to contribute ?

Details on the github page: [https://github.com/smarie/python-pytest-steps](https://github.com/smarie/python-pytest-steps)
